// Lib Ver 002
var _ = function() {
    var tidyTreeLib = new PlugIn.Library(new Version("0.01"));

    tidyTreeLib.tidyLayout = function(selection, lstCycle) {
        // GENERIC FUNCTIONS ---------------------------------------------
        // (++) :: [a] -> [a] -> [a]
        const append = (xs, ys) => xs.concat(ys);

        // concat :: [[a]] -> [a] | [String] -> String
        const concat = xs =>
            xs.length > 0 ? (() => {
                const unit = typeof xs[0] === 'string' ? '' : [];
                return unit.concat.apply(unit, xs);
            })() : [];

        // concatMap :: (a -> [b]) -> [a] -> [b]
        const concatMap = (f, xs) => [].concat.apply([], xs.map(f));

        // cons :: a -> [a] -> [a]
        const cons = (x, xs) => [x].concat(xs);

        // 2 or more arguments
        // curry :: Function -> Function
        const curry = (f, ...args) => {
            const go = xs => xs.length >= f.length ? (f.apply(null, xs)) :
                function() {
                    return go(xs.concat(Array.from(arguments)));
                };
            return go([].slice.call(args, 1));
        };

        //  (\\)  :: (Eq a) => [a] -> [a] -> [a]
        const difference = (xs, ys) =>
            xs.filter(x => ys.indexOf(x) === -1);

        // elem :: Eq a => a -> [a] -> Bool
        const elem = (x, xs) => xs.indexOf(x) !== -1;

        // elemIndex :: Eq a => a -> [a] -> Maybe Int
        const elemIndex = (x, xs) => {
            const i = xs.indexOf(x);
            return {
                nothing: i === -1,
                just: i
            };
        };

        // filter :: (a -> Bool) -> [a] -> [a]
        const filter = (f, xs) => xs.filter(f);

        // foldl :: (b -> a -> b) -> b -> [a] -> b
        const foldl = (f, a, xs) => xs.reduce(f, a);

        // fst :: [a, b] -> a
        const fst = pair => pair.length === 2 ? pair[0] : undefined;

        // id :: a -> a
        const id = x => x;

        // isNull :: [a] | String -> Bool
        const isNull = xs =>
            Array.isArray(xs) || typeof xs === 'string' ? (
                xs.length < 1
            ) : undefined;

        // intersectBy::(a - > a - > Bool) - > [a] - > [a] - > [a]
        const intersectBy = (eq, xs, ys) =>
            (xs.length > 0 && ys.length > 0) ?
            xs.filter(x => ys.some(curry(eq)(x))) : [];

        // log :: a -> IO ()
        const log = (...args) => {
            console.log(
                args
                .map(show)
                .join(' -> ')
            );
        };

        // map :: (a -> b) -> [a] -> [b]
        const map = (f, xs) => xs.map(f);

        // max :: Ord a => a -> a -> a
        const max = (a, b) => b > a ? b : a;

        // maximumBy :: (a -> a -> Ordering) -> [a] -> a
        const maximumBy = (f, xs) =>
            xs.reduce((a, x) => a === undefined ? x : (
                f(x, a) > 0 ? x : a
            ), undefined);

        // min :: Ord a => a -> a -> a
        const min = (a, b) => b < a ? b : a;

        // negate :: Num a => a -> a
        const negate = n => -n;

        // nextInCycleMay :: [a] -> a -> Maybe a
        const nextInCycleMay = (xs, x) => {
            const mbIndex = elemIndex(x, xs);
            return mbIndex.nothing ? (
                isNull(xs) ? mbIndex : {
                    just: xs[0]
                }
            ) : ({
                just: takeCycle(mbIndex.just + 2, xs)[mbIndex.just + 1],
                nothing: false
            });
        };

        // read :: Read a => String -> a
        const read = s => JSON.parse(s);

        // replicate :: Int -> a -> [a]
        const replicate = (n, x) =>
            Array.from({
                length: n
            }, () => x);

        // reverse :: [a] -> [a]
        const reverse = xs =>
            typeof xs === 'string' ? (
                xs.split('')
                .reverse()
                .join('')
            ) : xs.slice(0)
            .reverse();

        // show :: a -> String
        const show = (...x) =>
            JSON.stringify.apply(
                null, x.length > 1 ? [x[0], null, x[1]] : x
            );

        // snd :: (a, b) -> b
        const snd = tpl => Array.isArray(tpl) ? tpl[1] : undefined;

        // take :: Int -> [a] -> [a]
        const take = (n, xs) => xs.slice(0, n);

        // First n members of a repeating cycle of xs
        // takeCycle :: Int -> [a] -> [a]
        const takeCycle = (n, xs) => {
            const lng = xs.length;
            return take(n,
                (lng >= n ? xs : concat(replicate(Math.ceil(n / lng), xs)))
            );
        };

        // unconsMay :: [a] -> Maybe (a, [a])
        const unconsMay = xs => xs.length > 0 ? {
            just: [xs[0], xs.slice(1)],
            nothing: false
        } : {
            nothing: true
        };

        // unlines :: [String] -> String
        const unlines = xs => xs.join('\n');

        // unzip :: [(a,b)] -> ([a],[b])
        const unzip = xys =>
            xys.reduceRight(([xs, ys], [x, y]) => [
                [x].concat(xs), [y].concat(ys)
            ], [
                [],
                []
            ]);

        // zipWith :: (a -> b -> c) -> [a] -> [b] -> [c]
        const zipWith = (f, xs, ys) =>
            Array.from({
                length: Math.min(xs.length, ys.length)
            }, (_, i) => f(xs[i], ys[i]));

        // TEXT NEST FUNCTIONS -------------------------------------------

        // textNest :: OGoutlineNodes -> Node {text: String, nest:[Node]}
        const textNest = xs => {
            const go = (xs, iLevel) =>
                xs.length ? xs.map(x => {
                    const
                        g = x.graphic,
                        rect = g.geometry;
                    return {
                        id: g.id,
                        level: iLevel,
                        text: g.text,
                        xywh: ['x', 'y', 'width', 'height']
                            .reduce((a, k) => (a[k] = rect[k], a), {}),
                        nest: x.children.length > 0 ? (
                            go(x.children, iLevel + 1)
                        ) : []
                    };
                }) : [];
            return go(xs, 1);
        };

        // Level order iteration, collecting values by supplied keys

        // levelOrderTraversal :: [String] -> Node -> [[a]]
        const levelOrderTraversal = (ks, dctNode) => {
            const go = queue => {
                const mbht = unconsMay(queue);
                return mbht.nothing ? [] : (() => {
                    const [h, t] = mbht.just;
                    return cons(
                        foldl((a, k) => (a[k] = h[k], a), {}, ks),
                        go(isNull(h.nest) ? (
                            t
                        ) : append(t, h.nest)) // Nest deferred to later.
                    );
                })();
            };
            return go([dctNode]);
        };

        // levelHeights :: Bool -> Node -> Dictionary
        const levelHeights = (blnHoriz, dctNode) => {
            const k = blnHoriz ? 'width' : 'height';
            return foldl((a, x) => {
                const level = x.level;
                return (a[level] = max(a[level] || 0, x.xywh[k]), a)
            }, {}, levelOrderTraversal(['level', 'xywh'], dctNode));
        };

        // CANVAS --------------------------------------------------------
        const cnv = selection.canvas;

        // LAYOUT DIRECTION AND GAP SETTINGS -----------------------------

        // Rising Sort
        // [object HierarchicalDirection: Bottom]
        // [object HierarchicalDirection: Left]
        // [object HierarchicalDirection: Right]
        // [object HierarchicalDirection: Top]

        // layoutSettings :: OG () ->
        //      {direction: String, levelGap: Real, peerGap: Real}
        const layoutSettings = lstCycle => {
            const cnv = selection.canvas; //??? remove
            // directionName :: OG HierarchicalDirection -> String
            directionName = d =>
                d < HierarchicalDirection['Right'] ? (
                    d < HierarchicalDirection['Left'] ? (
                        'Bottom'
                    ) : 'Left'
                ) : d < HierarchicalDirection['Top'] ? (
                    'Right'
                ) : 'Top',
                layout = cnv.layoutInfo;

            return (
                layout.automaticLayout = false, // Lest it undo ours.
                layout.type = LayoutType.Hierarchical, // Remove ???
                // DIRECTION CYCLING FOR TESTING --- /// REMOVE ???

                lstCycle.length > 0 && (layout.direction =
                    HierarchicalDirection[nextInCycleMay(
                            lstCycle,
                            directionName(layout.direction))
                        .just]), {
                    direction: directionName(layout.direction),
                    levelGap: layout.rankSeparation * 72, // Points per inch.
                    peerGap: layout.objectSeparation * 72,
                    type: layout.type,
                    automatic: layout.automaticLayout
                });
        };

        // A positioned tree with its root at zero.
        // (A standard textNest decorated with geometry, id, level index)

        // layoutDesign :: Node {text: String, nest:[Node]} ->
        //           Node {text: String, nest:[Node], x:Real}
        const layoutDesign = (keyDimension, peerGap, dctNode) => {

            // REINGOLD TILFORD CODE ADAPTED FROM KENNEDY ----------------

            // TREE-DRAWING FUNCTIONS ------------------------------------
            // A JavaScript adaption of Andrew Kennedy's classic (ML) paper
            // 'Drawing Trees'
            // Journal of Functional Programming 6 (3): 527-534, May 1996

            // This version mainly differs from Kennedy in allowing for nodes
            // of variable height and width.
            // keyDimension is either 'width' (or vertical trees)
            // or 'height' for horizontal trees.
            // peerGap should be in the same units as height/width

            // X POSITIONS - relative to parents
            // The x position of a node is relative to its parent
            // (y positions are properties of nesting layers - ignored here)

            // A sub-tree moves with any x-delta applied to its parent node

            // Node {text: String, nest:[Node], x:Real}
            // moveTree :: Real -> Node -> Node
            const moveTree = (rd, dctNode) =>
                Object.assign({}, dctNode, {
                    rx: dctNode.rx + rd
                });

            // EXTENTS - Lists of pairs of x positions

            // The head of the list is the pair for the top level of the tree,
            // and thus downward.

            // Each pair holds the left (mininum) and right (maximum) x values
            // for its level
            // Extent [(leftMost :: Real, rightMost :: Real)]

            // Applying an x-axis delta to all the pairs in an extent:

            // moveExtent :: Real -> Extent -> Extent
            const moveExtent = (rd, pairs) =>
                map(([p, q]) => [p + rd, q + rd], pairs);

            // The left most positions of one extent,
            // combined with the rightmost positions of an adject extent:

            // merge :: Extent -> Extent -> Extent
            const merge = (ps, qs) => {
                const
                    mbp = unconsMay(ps),
                    mbq = mbp.nothing ? mbp : unconsMay(qs);
                return mbp.nothing ? (
                    qs
                ) : mbq.nothing ? (
                    ps
                ) : (() => {
                    const [ph, pt] = mbp.just, [qh, qt] = mbq.just;
                    return cons([fst(ph), snd(qh)], merge(pt, qt));
                })();
            };

            // Merging a pair of adjacent extents lifting to merging
            // a longer series of adjacent extents:

            // mergeList :: [Extent] -> Extent
            const mergeList = es => foldl(merge, [], es);

            // What is the *minimum* distance between two given extents ?
            // (Proliferating descendants may push them further apart)

            // fit :: Extent -> Extent -> Real
            const fit = (ps, qs) => {
                const
                    mbp = unconsMay(ps),
                    mbq = mbp.nothing ? mbp : unconsMay(qs);
                return (mbp.nothing || mbq.nothing) ? 0 : (() => {
                    const [ph, pt] = mbp.just, [qh, qt] = mbq.just;
                    return max((snd(ph) - fst(qh)), fit(pt, qt));
                })();
            };

            // How far leftward could each extent go ?

            // fitlistl :: [Extent] -> [Real]
            const fitlistl = es => {
                const fl = (a, ees) => {
                    const mbees = unconsMay(ees);
                    return mbees.nothing ? [] : (() => {
                        const [e, es] = mbees.just,
                            d = fit(a, e);
                        return cons(d, fl(merge(a, moveExtent(d, e)), es));
                    })();
                };
                return fl([], es);
            };

            // How far rightward could each extent go ?

            // fitlistr :: [Extent] -> [Real]
            const fitlistr = es => {
                const fr = (a, ees) => {
                    const mbees = unconsMay(ees);
                    return mbees.nothing ? [] : (() => {
                        const [e, es] = mbees.just,
                            d = negate(fit(e, a));
                        return cons(d, fr(merge(moveExtent(d, e), a), es));
                    })();
                }
                return reverse(fr([], reverse(es)));
            };

            // What are the mean positions,
            // mid way between left and right constraints,
            // for each extent ?

            // fitList :: [Extent] -> [Real]
            const fitList = es =>
                zipWith(
                    (x, y) => (x + y) / 2,
                    fitlistl(es),
                    fitlistr(es)
                );

            // design :: Node -> (Node, [Extent])
            const design = node => {
                const
                    w2 = (node.xywh[keyDimension] + peerGap) / 2,
                    [trees, extents] = unzip(
                        map(design, node.nest)
                    ),
                    positions = fitList(extents),
                    resultExtent = cons([-w2, w2], mergeList(
                        zipWith(moveExtent, positions, extents)
                    )),
                    resultTree = Object.assign({}, node, {
                        nest: zipWith(moveTree, positions, trees),
                        rx: 0
                    });
                return [resultTree, resultExtent];
            };
            return dctNode.nest.length > 0 ? fst(design(dctNode)) : dctNode;
        };

        // UPDATED RENDER ------------------------------------------------

        // Subtree positions updated from geometry in a given key of nodes
        // positionUpdated :: Canvas -> String -> Bool, Node -> Node
        const positionsUpdated = curry((cnv, key, blnAcross, dctNode) => {
            const
                dct = dctNode[key],
                [x, y, width, height] = map(
                    k => dct[k], ['x', 'y', 'width', 'height']
                ),
                g = cnv.graphicWithId(dctNode.id),
                mgs = g.magnets;

            // Make sure that each node has the right two magnets for
            // direction of tree growth. (NS for vertical, EW for horizontal)
            return (
                (g.magnets = (blnAcross ? (
                    [new Point(1.00, 0.00), new Point(-1.00, 0.00)]
                ) : [new Point(0.00, 1.00), new Point(0.00, -1.00)])),
                g.geometry = new Rect(x, y, width, height),
                map(positionsUpdated(cnv, key, blnAcross), dctNode.nest),
                dctNode
            );
        });

        // newPositions :: Dictionary -> Dictionary ->
        //                                  Dictionary -> Node -> Node
        const newPositions = (settings, dctLevels, posn, dctNode) => {
            // Direction, levelGap, peerGap
            const
                blnVert = elem(settings.direction, ['Top', 'Bottom']),
                blnMarked = elem(settings.direction, ['Bottom', 'Right']),
                [keyOrth, keyAxial, dimOrth, dimAxial] = blnVert ? (
                    ['width', 'height', 'x', 'y']
                ) : ['height', 'width', 'y', 'x'],
                cs = dctNode.nest,
                iLevel = dctNode.level,
                rankDepth = dctLevels[iLevel];
            return {
                id: dctNode.id,
                level: iLevel,
                text: dctNode.text,
                rx: dctNode.rx,
                geo: posn,
                nest: cs.length < 1 ? [] : (() => {
                    const
                        origin = posn[dimOrth] + (posn[keyOrth] / 2),
                        pLevelTop = blnMarked ? (
                            ((posn[dimAxial] + posn[keyAxial]) - rankDepth) -
                            settings.levelGap
                        ) : posn[dimAxial] + rankDepth + settings.levelGap;
                    return map(c => {
                        // Updated geo and recursively updated descendants.
                        const
                            cxywh = c.xywh,
                            w = (origin + c.rx) - (cxywh[keyOrth] / 2),
                            h = pLevelTop - (
                                blnMarked ? cxywh[keyAxial] : 0
                            );
                        return newPositions(settings, dctLevels, {
                            x: blnVert ? w : h,
                            y: blnVert ? h : w,
                            rx: c.rx,
                            width: cxywh.width,
                            height: cxywh.height
                        }, c);
                    }, cs)
                })()
            };
        };

        // Canvas -> List of graphics -> Unique list of their tree roots
        // allAncestors :: Canvas -> [Graphic] -> [Graphic]
        const allAncestors = (cnv, gs) =>
            map(x => cnv.graphicWithId(x),
                foldl((a, treeIDs) => {
                    const
                        lstNodes = a.nodes,
                        lstRest = difference(lstNodes, treeIDs);
                    return {
                        nodes: lstRest,
                        trees: lstRest.length < lstNodes.length ? (
                            cons(treeIDs[0], a.trees)
                        ) : a.trees
                    };
                }, {
                    nodes: map(g => g.id, gs),
                    trees: []
                }, map(xs => map(x => x.id, xs),
                    map(flatTree,
                        filter(
                            x => x.children.length > 0,
                            cnv.outlineRoot.children
                        ))))
                .trees
            );

        // flatTree :: OGOutlineNode -> [Graphic]
        const flatTree = oNode =>
            cons(oNode.graphic, concatMap(flatTree, oNode.children));

        // MAIN --------------------------------------------------------------

        // Trees only - childless top-level shapes need no layout
        // lstRoots :: [Node]
        const lstRoots = textNest(
            filter(x => x.children.length > 0, cnv.outlineRoot.children)
        );

        return (lstRoots.length < 1) ? [] : (() => {
            const
                // If any nodes selected, rotate only their trees (else all).
                lstSelnRoots = intersectBy(
                    (a, b) => a.id === b.id,
                    lstRoots, allAncestors(cnv, selection.graphics),
                ),
                // TEST VERSION ROTATES DIRECTION ???
                settings = layoutSettings(lstCycle),
                blnHoriz = elem(settings.direction, ['Left', 'Right']);

            return unlines(
                map(
                    dctNode => {
                        // New geometry,
                        const
                            blnTree = dctNode.nest.length > 0,
                            jsoLayout = blnTree ? layoutDesign(
                                blnHoriz ? 'height' : 'width',
                                settings.peerGap, dctNode
                            ) : {};
                        // newly rendered.
                        return show(
                            positionsUpdated(
                                cnv, 'geo', blnHoriz, newPositions(
                                    settings,
                                    levelHeights(blnHoriz, jsoLayout),
                                    jsoLayout.xywh,
                                    jsoLayout
                                )
                            ), 2
                        );
                    },
                    lstSelnRoots.length > 0 ? (
                        lstSelnRoots
                    ) : (isNull(lstRoots) ? (
                        [{
                            text: 'Root',
                            nest: []
                        }]
                    ) : lstRoots)
                )
            );
        })();
    };

    return tidyTreeLib;
}();
_;